import django_filters
from crispy_forms.bootstrap import InlineRadios
from crispy_forms.layout import Layout, Row, Div, Fieldset, Field
from django.contrib.contenttypes.models import ContentType
from django.forms import HiddenInput
from django.utils.translation import ugettext_lazy as _
from django_select2.forms import Select2Widget
from django.contrib.auth import get_user_model
from outputs.models import Export, Scheduler
from pragmatic.forms import SingleSubmitFormHelper
from pragmatic.filters import SliderFilter


class ExportFilter(django_filters.FilterSet):
    created = django_filters.DateFromToRangeFilter()
    total = SliderFilter(label=_('Total items'), step=10, has_range=True, segment='outputs.Export.total')
    creator = django_filters.ModelChoiceFilter(queryset=get_user_model().objects.all(), widget=Select2Widget)
    content_type = django_filters.ModelChoiceFilter(
        queryset=ContentType.objects.filter(pk__in=Export.objects.order_by('content_type').values_list('content_type', flat=True).distinct()),
        widget=Select2Widget
    )

    class Meta:
        model = Export
        fields = [
            'id', 'created', 'format', 'context', 'content_type', 'status',
            'total', 'creator'
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.fields['format'].empty_label = _("Doesn't matter")
        self.form.fields['context'].empty_label = _("Doesn't matter")
        self.helper = SingleSubmitFormHelper()
        self.helper.form_tag = False
        self.helper.disable_csrf = True
        self.helper.layout = Layout(
            Row(
                Div(
                    Row(
                        Fieldset(
                            _('Generic'),
                            Row(
                                Div('id', css_class='col-md-4'),
                                Div(Field('created', css_class='date-picker form-control'), css_class='col-md-8 range-filter')
                            ),
                            css_class='col-md-12'
                        ),
                    ),
                    Row(
                        Fieldset(
                            _('Related'),
                            Row(
                                Div('creator', css_class='col-md-5'),
                                Div(Field('total', css_class='form-control'), css_class='col-md-7 range-filter')
                            ),
                            Row(
                                Div('content_type', css_class='col-md-5')
                            ),
                            css_class='col-md-12'
                        ),
                    ),
                    css_class='col-md-9'
                ),
                Fieldset(
                    _('Attributes'),
                    InlineRadios('format'),
                    InlineRadios('context'),
                    css_class='col-md-3'
                ),
            )
        )


class ExportFilterSet(django_filters.FilterSet):
    export = django_filters.ModelChoiceFilter(queryset=Export.objects.all(), widget=HiddenInput(), label=_('Export'), method='filter_export')

    def filter_export(self, queryset, name, value):
        return queryset.filter(id__in=value.object_list)


class SchedulerFilter(django_filters.FilterSet):
    created = django_filters.DateFromToRangeFilter()
    creator = django_filters.ModelChoiceFilter(queryset=get_user_model().objects.all(), widget=Select2Widget)
    content_type = django_filters.ModelChoiceFilter(
        queryset=ContentType.objects.filter(pk__in=Scheduler.objects.order_by('content_type').values_list('content_type', flat=True).distinct()),
        widget=Select2Widget
    )
    is_active = django_filters.ChoiceFilter(label=_('Active'), empty_label=_("Doesn't matter"), choices=[('True', _("Yes")), ('False', _("No"))])

    class Meta:
        model = Scheduler
        fields = [
            'id', 'created', 'format', 'context', 'content_type',
            'creator', 'routine', 'is_active'
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.fields['format'].empty_label = _("Doesn't matter")
        self.form.fields['context'].empty_label = _("Doesn't matter")
        self.helper = SingleSubmitFormHelper()
        self.helper.form_tag = False
        self.helper.disable_csrf = True
        self.helper.layout = Layout(
            Row(
                Div(
                    Row(
                        Fieldset(
                            _('Generic'),
                            Row(
                                Div('id', css_class='col-md-4'),
                                Div(Field('created', css_class='date-picker form-control'), css_class='col-md-8 range-filter')
                            ),
                            css_class='col-md-12'
                        ),
                    ),
                    Row(
                        Fieldset(
                            _('Related'),
                            Row(
                                Div('creator', css_class='col-md-5'),
                                Div(Field('total', css_class='form-control'), css_class='col-md-7 range-filter')
                            ),
                            Row(
                                Div('content_type', css_class='col-md-5')
                            ),
                            css_class='col-md-12'
                        ),
                    ),
                    css_class='col-md-9'
                ),
                Fieldset(
                    _('Attributes'),
                    InlineRadios('format'),
                    InlineRadios('context'),
                    InlineRadios('routine'),
                    InlineRadios('is_active'),
                    css_class='col-md-3'
                ),
            )
        )



from django.contrib.auth import get_user_model
from pragmatic.signals import apm_custom_context


@apm_custom_context('tasks')
def notify_about_executed_export(export):
    from whistle.helpers import notify
    notify_users = get_user_model().objects \
        .filter(is_active=True, is_superuser=True) \
        .exclude(pk=export.creator.pk) \
        .exclude(pk__in=export.recipients.all())

    for user in notify_users:
        notify(None, user, 'EXPORT_EXECUTED', actor=export.creator, object=export, target=export.content_type)


@apm_custom_context('tasks')
def schedule_scheduler(scheduler):
    scheduler.schedule()



# Generated by Django 3.1.14 on 2023-01-25 11:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0012_auto_20230125_1252'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='scheduler',
            name='custom_cron_string',
        ),
        migrations.AddConstraint(
            model_name='scheduler',
            constraint=models.CheckConstraint(check=models.Q(models.Q(('cron_string', ''), models.Q(_negated=True, routine='CUSTOM')), models.Q(models.Q(_negated=True, cron_string=''), ('routine', 'CUSTOM')), _connector='OR'), name='custom_cron_string'),
        ),
    ]



from django.urls import path
from django.utils.translation import pgettext_lazy

from outputs.views import ExportListView, SchedulerListView, SchedulerCreateView, SchedulerUpdateView, SchedulerDetailView, SchedulerDeleteView

app_name = 'outputs'

urlpatterns = [
    path(pgettext_lazy('url', 'exports/'), ExportListView.as_view(), name='export_list'),
    path(pgettext_lazy("url", 'schedulers/<int:pk>/'), SchedulerDetailView.as_view(), name='scheduler_detail'),
    path(pgettext_lazy("url", 'schedulers/<int:pk>/update/'), SchedulerUpdateView.as_view(), name='scheduler_update'),
    path(pgettext_lazy("url", 'schedulers/<int:pk>/delete/'), SchedulerDeleteView.as_view(), name='scheduler_delete'),
    path(pgettext_lazy("url", 'schedulers/create/from-export/<int:export_pk>/'), SchedulerCreateView.as_view(), name='scheduler_create_from_export'),
    path(pgettext_lazy("url", 'schedulers/create/'), SchedulerCreateView.as_view(), name='scheduler_create'),
    path(pgettext_lazy('url', 'schedulers/'), SchedulerListView.as_view(), name='scheduler_list'),
]



# Generated by Django 3.1.14 on 2023-01-11 12:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0008_auto_20230111_1350'),
    ]

    operations = [
        migrations.AddField(
            model_name='export',
            name='exporter_path',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
        migrations.AddField(
            model_name='scheduler',
            name='exporter_path',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
    ]



import json

from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import migrations
from django.utils.module_loading import import_string

from outputs.models import AbstractExport


def reformat_export_fields_permissions(*args, **kwargs):
    for user in get_user_model().objects.exclude(export_fields_permissions={}):
        permissions = user.export_fields_permissions
        new_permissions = {}

        # {'invoicing.Invoice.Xlsx': ['id', 'created', 'get_type_dis...
        if isinstance(permissions, str):
            permissions = json.loads(permissions)

        for key, fields in permissions.items():
            app_label, model_name, format = key.split('.')
            model_class = apps.get_model(app_label=app_label, model_name=model_name)
            exporter_path = AbstractExport.get_exporter_path(model_class, AbstractExport.CONTEXT_LIST, format)
            exporter = import_string(exporter_path)
            new_permissions[exporter.get_path()] = fields

        user.export_fields_permissions = new_permissions
        user.save(update_fields=['export_fields_permissions'])


def undo_reformat_export_fields_permissions(*args, **kwargs):
    for user in get_user_model().objects.exclude(export_fields_permissions={}):
        permissions = user.export_fields_permissions
        old_permissions = {}

        # {'swida.core.accounts.exporters.TeamXlsxListExporter': ['id',
        if isinstance(permissions, str):
            permissions = json.loads(permissions)

        for key, fields in permissions.items():
            exporter = import_string(key)
            app_label, model_name = exporter.get_app_and_model()
            export_format = exporter.export_format
            export_format = export_format.capitalize()
            old_permission = '.'.join([app_label, model_name, export_format])
            old_permissions[old_permission] = fields

        user.export_fields_permissions = old_permissions
        user.save(update_fields=['export_fields_permissions'])


class Migration(migrations.Migration):
    dependencies = [
        ('outputs', '0010_set_exporter_path'),
    ]

    operations = [
        migrations.RunPython(reformat_export_fields_permissions, undo_reformat_export_fields_permissions),
    ]



# Generated by Django 2.2.9 on 2020-08-24 12:18

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0003_init_emails'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='scheduler',
            name='emails',
        ),
        migrations.AlterField(
            model_name='export',
            name='emails',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.EmailField(max_length=254), default=list, size=None),
        ),
    ]



import json
from collections import OrderedDict

from crispy_forms.layout import Field, Div
from crispy_forms.utils import TEMPLATE_PACK
from django.core.exceptions import ValidationError
from django.db.models import QuerySet
from django.forms import CheckboxSelectMultiple, MultipleChoiceField
from django.forms.widgets import ChoiceWidget
from django.template.loader import get_template
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe

from outputs import settings as outputs_settings

try:
    # Django 3.1
    from django.db.models import JSONField
except ImportError:
    # older Django
    from django.contrib.postgres.fields import JSONField


class ExportFieldsPermissionsMixin(object):
    def get_choices(self):
        if not self.choices:
            self.choices = self.load_choices()
        return self.choices

    def get_all_exportable_fields(self):
        if not hasattr(self, 'all_exportable_fields'):
            self.load_all_exportable_fields()
        return self.all_exportable_fields

    def get_max_field_groups(self):
        if not hasattr(self, 'max_field_groups'):
            self.load_all_exportable_fields()
        return self.max_field_groups

    def get_table(self):
        if not hasattr(self, 'table'):
            self.load_table_and_width()
        return self.table

    def get_table_width(self):
        if not hasattr(self, 'table'):
            self.load_table_and_width()
        return self.table_width

    def load_choices(self):
        choices = []

        # we are starting with dictionary to guarantee ordered vertical structure of the table
        table = self.get_table()

        for row in table:
            choices.append(row['exporter_path'])

            for group in row['field_groups']:
                # row may contain empty groups to fill up empty columns, skipping those here
                if group:
                    choices.append(group['key'])

                    for permission in group['permissions']:
                        choices.append(permission['key'])

        return choices

    def load_table_and_width(self):
        table = []

        # we are starting with dictionary to guarantee ordered vertical structure of the table
        exportable_fields = self.get_all_exportable_fields()
        max_field_groups = self.get_max_field_groups()

        for exporter_path, field_groups in exportable_fields.items():
            # TODO: not every exporter subclass contains model
            exporter = import_string(exporter_path)
            app_label, model_name = exporter.get_app_and_model()

            row = {
                'app': app_label,
                'model': model_name,
                'format': exporter.export_format.capitalize(),
                'description': exporter.get_description(),
                'exporter_path': exporter_path,
                'field_groups': []
            }

            index = 0
            for group, fields in field_groups.items():
                group_dict = {
                    'label': group,
                    'key': '/'.join([row['exporter_path'], 'group', str(index)]),
                    'permissions': []
                }

                for field in fields:
                    group_dict['permissions'].append({
                        'label': field[1],
                        'key': '/'.join([row['exporter_path'], field[0]]),
                    })

                row['field_groups'].append(dict(group_dict))
                index += 1

            # fill the row with empty groups to have equal number of columns in every row
            if index < max_field_groups:
                for i in range(index, max_field_groups):
                    row['field_groups'].append({})

            table.append(row)

        # TODO: sort by x['app'] and x['model']
        self.table = sorted(table, key=lambda x: x['app'])
        self.table_width = max_field_groups+1

    def load_all_exportable_fields(self):
        exporters = set()

        from outputs.mixins import ExcelExporterMixin
        for cls in ExcelExporterMixin.__subclasses__():
            if hasattr(cls, 'selectable_fields') and \
                not cls.exclude_in_permission_widget and \
                cls.get_path() not in outputs_settings.EXCLUDE_EXPORTERS:
                exporters.add(cls)

        all_exportable_fields = {}
        max_field_groups = 0

        for exporter in exporters:
            try:
                selectable_fields = exporter.selectable_fields()
            except AttributeError:
                selectable_fields = {}

            # enforcing order of dict in case it was unordered as it is crusial for the later use in table
            if not isinstance(selectable_fields, OrderedDict):
                selectable_fields = OrderedDict(selectable_fields)

            try:
                for iterative_set in exporter.selectable_iterative_sets().values():
                    selectable_fields.update(iterative_set)
            except AttributeError:
                pass

            all_exportable_fields[exporter.get_path()] = selectable_fields

            # looking for max number of groups which later translates to number of table columns
            if len(selectable_fields) > max_field_groups:
                max_field_groups = len(selectable_fields)

        self.all_exportable_fields = all_exportable_fields
        self.max_field_groups = max_field_groups

    def decompress(self, value):
        """
        Decompress value (expected to be string of json dumped dictionary) to list of initial values for template table
        """
        # is single value passed, convert to tuple, else raise error if not anticipated iterable
        if isinstance(value, str) or isinstance(value, dict):
            value = (value,)
        elif not isinstance(value, (list, tuple, QuerySet)):
            raise TypeError()

        exportable_fields = self.get_all_exportable_fields()
        permission_keys = set()

        for val in value:
            permissions = json.loads(val) if isinstance(val, str) else val
            # when val is retrieved from queryset, load needs to be done twice for some reason unknown to me
            if isinstance(permissions, str):
                permissions = json.loads(permissions)

            # due to previous bug this case is handled instead of creating migration
            if permissions is None:
                permissions = {}
            # if two loads where not enough there gotta be something wrong
            elif not isinstance(permissions, dict):
                raise TypeError()

            for exporter_path, permitted_fields in permissions.items():
                for field in permitted_fields:
                    permission_keys.add('/'.join([exporter_path, field]))

                # set group and exporter fields
                exporter_fields_count = 0
                group_index = -1

                for group, group_fields in exportable_fields[exporter_path].items():
                    exporter_fields_count += len(group_fields)
                    group_index += 1
                    group_permitted_count = 0

                    for field in group_fields:
                        if field[0] not in permitted_fields:
                            break
                        group_permitted_count += 1

                    # if all fields within a group are permitted add group_key too (to initial values)
                    if group_permitted_count == len(group_fields):
                        permission_keys.add('/'.join([exporter_path, 'group', str(group_index)]))

                # if all fields of exporter are permitted add exporter_path too (to initial values)
                if len(permitted_fields) == exporter_fields_count:
                    permission_keys.add(exporter_path)

        return permission_keys

    def format_output(self, value, compress=True):
        """
        Converts value (list) to dictionary and optionally compress to json string
        """
        if not value:
            return {}

        output = {}
        for val in value:
            keys = val.split('/')
            # key length is supposed to be 1 for exporter, 2 for field, 3 for group
            if len(keys) > 3:
                raise ValueError()
            # we output only field values, exporter and group keys are only auxiliary
            elif len(keys) == 2:
                exporter_path = keys[0]
                field = keys[1]

                if exporter_path not in output:
                    output[exporter_path] = []

                output[exporter_path].append(field)

        # output dictionary compressed to string if compress=True and it is not empty
        if output and compress:
            return json.dumps(output)
        else:
            return output


class ExportFieldsPermissionsSelectMultipleWidget(CheckboxSelectMultiple, ExportFieldsPermissionsMixin):
    template = 'outputs/widgets/export_fields_permissions_widget.html'

    def __init__(self, attrs=None):
        super().__init__(attrs)

    def render(self, name, value, attrs=None, choices=(), renderer=None):
        t = get_template(self.template)

        # decompress value (json compressed dictionary) into list if exists, else pass empty tuple
        if value:
            value = self.decompress(value)
        else:
            value = ()

        c = {
            'name': name,
            'value': value,
            'table': self.get_table(),
            'width': self.get_table_width(),
        }

        return mark_safe(t.render(c))


class ExportFieldsPermissionsField(MultipleChoiceField):
    """
    This Field is intended to be used with ExportFieldsPermissionsSelectMultipleWidget,
    without it ExportFieldsPermissionsSelectMultipleWidget's render method needs to be implemented
    """
    widget = ExportFieldsPermissionsSelectMultipleWidget

    def __init__(self, *args, **kwargs):
        # widget.get_choices can't be used because exporters in widget.get_all_exportable_fields are not yet available
        super().__init__(*args, **kwargs)
        self.required=False

    def validate(self, value):
        """Validate that the input is a list or tuple."""
        if self.required and not value:
            raise ValidationError(self.error_messages['required'], code='required')

        if not isinstance(value, (list, tuple)):
            raise TypeError()

        # Validate that each value in the value list is in self.choices.
        for val in value:
            if val not in self.choices:
                raise ValidationError(
                    self.error_messages['invalid_choice'],
                    code='invalid_choice',
                    params={'value': val},
                )

    def clean(self, value):
        # get choices for validation, at this point exporters are already available
        self.choices = self.widget.get_choices()

        # clean using modified validate() method
        value = self.to_python(value)
        self.validate(value)
        self.run_validators(value)
        return self.widget.format_output(value)


class ExportFieldsPermissionsModelField(JSONField):
    """
    Custom export fields permissions field based on JSONField
    """
    form_class = ExportFieldsPermissionsField

    def formfield(self, **kwargs):
        kwargs.update({'widget': self.form_class.widget})
        field = self.form_class(**kwargs)
        return field


class ExportFieldsPermissionsCrispyField(Field):
    template = 'outputs/widgets/export_fields_permissions_layout.html'

    def __init__(self, *args, **kwargs):
        self.initial = kwargs.pop('initial', None)
        self.initial_group_permissions = kwargs.pop('initial_group', None)
        super().__init__(*args, **kwargs)
        self.widget = ExportFieldsPermissionsSelectMultipleWidget()

    def render(self, form, form_style, context, template_pack=TEMPLATE_PACK, extra_context=None, **kwargs):
        if extra_context is None:
            extra_context = {}

        extra_context['table'] = self.widget.get_table()
        extra_context['width'] = self.widget.get_table_width()

        # decompress initial value (json compressed dictionary) into list if exist, else pass empty tuple
        if self.initial:
            initial = self.widget.decompress(self.initial)
        else:
            initial = ()

        # decompress group permissions if there are any
        if self.initial_group_permissions:
            groups_permissions = self.widget.decompress(self.initial_group_permissions)
        else:
            groups_permissions = ()

        form.initial['export_fields_permissions'] = initial
        extra_context['groups_permissions'] = groups_permissions

        return super().render(form, form_style, context, template_pack, extra_context, **kwargs)


class CheckboxSelectMultipleWithDisabled(ChoiceWidget):
    """
    Subclass of Django's ChoiceWidget and based on ChcekboxSelectMultiple widget that allows disabling options according to https://djangosnippets.org/snippets/2453/.
    To disable an option, pass a dict instead of a string for its label,
    of the form: {'label': 'option label', 'disabled': True}
    """
    allow_multiple_selected = True
    input_type = 'checkbox'
    template_name = "outputs/widgets/custom_checkbox_select.html"
    option_template_name = 'django/forms/widgets/checkbox_option.html'

    def use_required_attribute(self, initial):
        # Don't use the 'required' attribute because browser validation would
        # require all checkboxes to be checked instead of at least one.
        return False

    def value_omitted_from_data(self, data, files, name):
        # HTML checkboxes don't appear in POST data if not checked, so it's
        # never known if the value is actually omitted.
        return False

    def id_for_label(self, id_, index=None):
        """"
        Don't include for="field_0" in <label> because clicking such a label
        would toggle the first checkbox.
        """
        if index is None:
            return ''
        return super().id_for_label(id_, index)

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        disabled = False

        # call parent method with label and supply with disabled attribute afterwatds
        if isinstance(label, dict):
            label, disabled = label['label'], label['disabled']

        option_dict = super().create_option(name, value, label, selected, index, subindex=subindex, attrs=attrs)

        if disabled:
            option_dict['attrs']['disabled'] = 'disabled'
        return option_dict


class Legend(Div):
    template = "outputs/widgets/legend.html"



from django.db import connection, migrations


def change_column_type(*args, **kwargs):
    with connection.cursor() as cursor:
        # changing type from character varying(16) to integer
        cursor.execute("ALTER TABLE public.outputs_export_items ALTER COLUMN gm2m_pk TYPE integer USING gm2m_pk::integer;")


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('outputs', '0004_auto_20200824_1418'),
    ]

    operations = [
        migrations.RunPython(change_column_type)
    ]



import inspect

from cron_descriptor import CasingTypeEnum, ExpressionDescriptor
from croniter import croniter
from django.core.exceptions import ValidationError
from django.utils.timezone import now

from auditlog.models import AuditlogHistoryField
from auditlog.registry import auditlog
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField
from django.core.validators import EMPTY_VALUES
from django.db import models
from django.http import QueryDict
from django.template import Context, Template
from django.template.defaultfilters import title
from django.urls import reverse, NoReverseMatch, resolve, Resolver404
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _, get_language
from gm2m import GM2MField
from pytz import timezone
from rq.exceptions import NoSuchJobError
from rq.job import Job

from outputs import jobs, settings as outputs_settings
from outputs.cron import schedule_export
from outputs.querysets import ExportQuerySet, SchedulerQuerySet

from pragmatic.templatetags.pragmatic_tags import filtered_values


exporters_module_mapping = outputs_settings.EXPORTERS_MODULE_MAPPING
related_models = outputs_settings.RELATED_MODELS


class AbstractExport(models.Model):
    FORMAT_XLSX = 'XLSX'
    FORMAT_XML_MRP = 'XML_MRP'
    FORMAT_PDF = 'PDF'
    FORMATS = [
        (FORMAT_XLSX, 'XLSX'),
        (FORMAT_XML_MRP, 'XML MRP'),
        (FORMAT_PDF, 'PDF'),
    ]

    CONTEXT_LIST = 'LIST'
    CONTEXT_STATISTICS = 'STATISTICS'
    CONTEXT_DETAIL = 'DETAIL'
    CONTEXTS = [
        (CONTEXT_LIST, _('list')),
        (CONTEXT_STATISTICS, _('statistics')),
        (CONTEXT_DETAIL, _('detail'))
    ]

    content_type = models.ForeignKey(ContentType, verbose_name=_('content type'), on_delete=models.CASCADE)
    format = models.CharField(_('format'), choices=FORMATS, max_length=7)
    context = models.CharField(_('context'), choices=CONTEXTS, max_length=10)
    exporter_path = models.CharField(_('exporter path'), max_length=100, blank=True, default='')
    fields = ArrayField(verbose_name=_('fields'), base_field=models.CharField(max_length=40), blank=True, null=True, default=None)
    query_string = models.TextField(_('query string'), blank=True, default='')
    creator = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_('creator'), on_delete=models.PROTECT, related_name="%(class)s_where_creator",
                                blank=True, null=True, default=None)
    recipients = models.ManyToManyField(get_user_model(), verbose_name=_('recipients'), related_name="%(class)s_where_recipient", blank=True)
    created = models.DateTimeField(_('created'), auto_now_add=True)
    modified = models.DateTimeField(_('modified'), auto_now=True)
    send_separately = models.BooleanField(_('send separately'), default=False)

    class Meta:
        abstract = True

    @property
    def model_class(self):
        return self.content_type.model_class()

    @staticmethod
    def get_exporter_path(model_class, context, format):
        """
        Returns generic exporter path based on content object model, export context and format.
        """
        exports_module = exporters_module_mapping.get(model_class._meta.label, None)

        if isinstance(exports_module, dict):
            exports_module = exports_module.get(context, None)

        if not exports_module:
            models_module = model_class.__module__
            app_module = models_module.rsplit('.', maxsplit=1)[0]
            exports_module = f'{app_module}.exporters'

        exporter_class_name = model_class.__name__
        format = title(format)
        format = format.replace('_', '')
        context = title(context)
        exporter_class_name = '{}{}{}Exporter'.format(exporter_class_name, format, context)
        exporter_path = f'{exports_module}.{exporter_class_name}'
        return exporter_path

    @property
    def exporter_class(self):
        try:
            if self.exporter_path:
                return import_string(self.exporter_path)
            else:
                raise ModuleNotFoundError()
        except ModuleNotFoundError:
            # TODO: log
            exporter_path = AbstractExport.get_exporter_path(
                model_class=self.model_class,
                context=self.context,
                format=self.format
            )
            print('ERROR: Exporter path %s not found. Using fallback (predicted path) instead: %s' % (self.exporter_path, exporter_path))
            return import_string(exporter_path)

    @property
    def exporter(self):
        constructor = self.exporter_class.__init__
        signature = inspect.signature(constructor)
        arguments = signature.parameters.keys()

        params = self.exporter_params.copy()

        for key in list(params.keys()):
            if key not in arguments and 'kwargs' not in arguments:
                params.pop(key)

        return self.exporter_class(**params)

    @property
    def exporter_params(self):
        return {
            'params': self.params,
            'user': self.creator,
            'recipients': self.recipients.all(),
            'selected_fields': self.fields,
            'send_separately': self.send_separately
        }

    @property
    def params(self):
        return QueryDict(self.query_string)

    @property
    def recipients_emails(self):
        return list(self.recipients.values_list('email', flat=True))

    def get_params_display(self):
        result = ''

        try:
            filter = self.exporter.filter
            values = filtered_values(filter, self.params)

            for param, field in values.items():
                label = field['label']
                value = field['value']

                value = Template("{{ value }}").render(Context({'value': value}))  # workaround for datetime ranges
                result = f'{result}{label}: {value}\n'
        except AttributeError:
            values = self.params

            for param, value in values.items():
                result = f'{result}{param}: {value}\n'

        return result

    def get_fields_labels(self):
        if self.fields is None:  # TODO: double check functionality
            return []

        field_labels = []
        exporter = self.exporter

        try:
            selectable_fields = exporter.selectable_fields()
        except AttributeError:
            selectable_fields = None

        try:
            for set in exporter.selectable_iterative_sets().values():
                selectable_fields.update(set)
        except AttributeError:
            pass

        selectable_fields_with_labels = {}
        if selectable_fields:
            for field_group in selectable_fields.values():
                for field in field_group:
                    selectable_fields_with_labels[field[0]] = field[1]

        for field in self.fields:
            field_labels.append(selectable_fields_with_labels.get(field, field))

        return field_labels


class Export(AbstractExport):
    STATUS_PENDING = 'PENDING'
    STATUS_PROCESSING = 'PROCESSING'
    STATUS_FAILED = 'FAILED'
    STATUS_FINISHED = 'FINISHED'
    STATUSES = [
        (STATUS_PENDING, _('pending')),
        (STATUS_PROCESSING, _('processing')),
        (STATUS_FAILED, _('failed')),
        (STATUS_FINISHED, _('finished'))
    ]
    status = models.CharField(_('status'), choices=STATUSES, max_length=10, default=STATUS_PENDING)
    items = GM2MField(*related_models, related_name='exports_where_item')
    total = models.PositiveIntegerField(_('total items'), default=0)
    emails = ArrayField(
        verbose_name=_('emails'),
        base_field=models.EmailField(),
        default=list,
    )
    objects = ExportQuerySet.as_manager()
    history = AuditlogHistoryField()

    class Meta:
        verbose_name = _('export')
        verbose_name_plural = _('exports')
        ordering = ('created',)
        default_permissions = getattr(settings, 'DEFAULT_PERMISSIONS', ('add', 'change', 'delete', 'view'))

    def __str__(self):
        model = self.content_type.model_class()
        name = model._meta.verbose_name_plural if model else self.content_type.model
        return '{} #{} ({})'.format(_('Export'), self.pk, name)

    def _get_base_url(self):
        try:
            app_label = self.get_app_label()
            url = reverse(f'{app_label}:{self.content_type.model}_list')
            return url
        except NoReverseMatch:
            model = self.content_type.model_class()
            instance = model(pk=1)
            detail_url = instance.get_absolute_url()

            from urllib.parse import urlsplit
            parsed = urlsplit(detail_url)
            try:
                match = resolve(parsed.path)
                to_be_reversed = "%s:%s" % (match.namespace, match.url_name) if match.namespace else match.url_name
                list_url = to_be_reversed.replace('detail', 'list')
                return reverse(list_url)
            except (Resolver404, NoReverseMatch):
                pass

        return None
        # return NotImplementedError()

    def get_items_url(self):
        base_url = self._get_base_url()
        return f'{self._get_base_url()}?export={self.pk}' if base_url is not None else None

    def get_absolute_url(self):
        base_url = self._get_base_url()
        return f'{self._get_base_url()}?{self.query_string}' if base_url is not None else None

    def get_app_label(self):
        if self.context in [self.CONTEXT_LIST, self.CONTEXT_DETAIL]:
            app_label = self.content_type.app_label

            if app_label == 'invoicing':  # TODO: refactor
                app_label = 'billing'
            elif app_label == 'exports':
                app_label = 'outputs'

        else:
            module = exporters_module_mapping[self.model_class._meta.label][self.context]
            app_label = module.split('.')[-2]

        return app_label

    def send_mail(self, language, filename=None):
        export_class_name = f'{self.__class__.__module__}.{self.__class__.__name__}'
        jobs.mail_export.delay(self.pk, export_class_name, language, filename)

    @property
    def object_list(self):
        ids = list(self.items.all().values_list('gm2m_pk', flat=True))
        # ids = list(map(int, ids))
        model = self.content_type.model_class()
        return model.objects.filter(id__in=ids)

    @property
    def exporter_params(self):
        return {
            'params': self.params,
            'items': self.object_list,  # this is required as we want to send identically same export (not currently available filtered data)
            'user': self.creator,
            'recipients': self.recipients.all(),
            'selected_fields': self.fields
        }


class Scheduler(AbstractExport):
    ROUTINE_OFTEN = 'OFTEN'                 # for debug purposes
    ROUTINE_DAILY = 'DAILY'                 # every morning at 8:00
    ROUTINE_WEEKLY = 'WEEKLY'               # every monday at 8:00
    ROUTINE_MONTHLY = 'MONTHLY'             # at 1st of current month TODO: reports will be for the previous month
    ROUTINE_CUSTOM = 'CUSTOM'               # custom cron string
    ROUTINES = [
        # (ROUTINE_OFTEN, _('often')),
        (ROUTINE_DAILY, _('daily')),
        (ROUTINE_WEEKLY, _('weekly')),
        (ROUTINE_MONTHLY, _('monthly')),
        (ROUTINE_CUSTOM, _('custom'))
    ]

    ROUTINE_DESCRIPTIONS = {
        # ROUTINE_OFTEN: _('every minute'),  # 6 UTC
        ROUTINE_DAILY: _('at 8:00'),  # 6 UTC
        ROUTINE_WEEKLY: _('on Monday'),
        ROUTINE_MONTHLY: _('on the first day'),
        ROUTINE_CUSTOM: _('defined by user'),
    }

    routine = models.CharField(_('routine'), choices=ROUTINES, max_length=7)
    cron_string = models.CharField(_('cron string'), max_length=13, blank=True)
    is_active = models.BooleanField(_('active'), default=True)
    executions = ArrayField(verbose_name=_('executions'), base_field=models.DateTimeField(), blank=True, default=list)
    job_id = models.CharField('job ID', max_length=36, blank=True)
    language = models.CharField(_('language'), choices=settings.LANGUAGES, max_length=2, db_index=True, default='en')
    # TODO: filename
    objects = SchedulerQuerySet.as_manager()
    history = AuditlogHistoryField()

    class Meta:
        verbose_name = _('scheduler')
        verbose_name_plural = _('schedulers')
        ordering = ('created',)
        default_permissions = getattr(settings, 'DEFAULT_PERMISSIONS', ('add', 'change', 'delete', 'view'))

        constraints = [
            # constraint of empty cront string for custom routine
            models.CheckConstraint(check=
                                   # models.Q(cron_string='') ^ ~models.Q(routine='CUSTOM'),  # Added in Django 4.1 (https://docs.djangoproject.com/en/4.1/releases/4.1/#models)
                                   models.Q(models.Q(cron_string='') & ~models.Q(routine='CUSTOM')) |
                                   models.Q(~models.Q(cron_string='') & models.Q(routine='CUSTOM')),
                                   name='custom_cron_string')
        ]

    def __str__(self):
        return '{} #{} ({} - {})'.format(_('Scheduler'), self.pk, self.content_type.model_class()._meta.verbose_name_plural, self.get_routine_display())

    def clean(self):
        if not((self.cron_string == '') ^ (self.routine == self.ROUTINE_CUSTOM)):
            if self.cron_string != '':
                raise ValidationError(_('Cron string has to be empty for routine %s') % self.get_routine_display())
            if self.routine == self.ROUTINE_CUSTOM:
                raise ValidationError(_('Missing cron string'))

        if self.routine == self.ROUTINE_CUSTOM:
            try:
                croniter(self.cron_string, now())
            except Exception as e:
                raise ValidationError(_('Invalid cron string: %s') % str(e))

        return super().clean()

    def get_absolute_url(self):
        return reverse('outputs:scheduler_detail', args=(self.pk,))

    @property
    def job(self):
        if self.job_id in EMPTY_VALUES:
            return None

        queue = django_rq.get_queue('cron')

        try:
            return Job.fetch(self.job_id, connection=queue.connection)
        except NoSuchJobError:
            return None

    @property
    def schedule_time(self):
        if not self.is_scheduled:
            return None

        # get all scheduled jobs
        scheduler = django_rq.get_scheduler('cron')
        jobs = scheduler.get_jobs(with_times=True)

        # get scheduler job by its ID
        # job, scheduled_at = list(filter(lambda x: x[0].id == self.job_id, jobs))[0]  # this is slower because it iterates whole array
        job, scheduled_at = next(x for x in jobs if x[0].id == self.job_id)            # returns first match

        # read time from scheduler
        scheduled_at = scheduled_at.replace(tzinfo=timezone('UTC'))
        scheduled_at = scheduled_at.astimezone(timezone(settings.TIME_ZONE))
        return scheduled_at

    @property
    def is_scheduled(self):
        return self.job is not None

    @property
    def routine_description(self):
        return self.ROUTINE_DESCRIPTIONS.get(self.routine)

    @property
    def cron_description(self):
        language = get_language()

        descriptor = ExpressionDescriptor(
            expression=self.get_cron_string(),
            throw_exception_on_parse_error=False,
            casing_type=CasingTypeEnum.Sentence,
            use_24hour_time_format=True,
            locale_code='%s_%s' % (language, language.upper())
        )
        return descriptor.get_description()

    def cancel_schedule(self):
        if self.is_scheduled:
            self.job.delete()

    def schedule(self):
        # cancel previous cron job
        self.cancel_schedule()

        if self.is_active:
            # schedule cronjob for active scheduler and save its ID
            scheduler = django_rq.get_scheduler('cron')

            # schedule export as cron job
            scheduler_class_name = f'{self.__class__.__module__}.{self.__class__.__name__}'

            try:
                job = scheduler.cron(
                    cron_string=self.get_cron_string(),
                    func=schedule_export,
                    args=(self.pk, scheduler_class_name),
                    timeout=settings.RQ_QUEUES['cron']['DEFAULT_TIMEOUT']
                )
                self.job_id = job.id
            except Exception as e:
                # TODO: log
                print(e)
        else:
            # inactive scheduler doesn't have job ID
            self.job_id = ''

        self.save(update_fields=['job_id'])

    def get_cron_string(self):
        # ┌───────────── minute (0 - 59)
        # │ ┌───────────── hour (0 - 23)
        # │ │ ┌───────────── day of the month (1 - 31)
        # │ │ │ ┌───────────── month (1 - 12)
        # │ │ │ │ ┌───────────── day of the week (0 - 6) (Sunday to Saturday;
        # │ │ │ │ │                                   7 is also Sunday on some systems)
        # │ │ │ │ │
        # │ │ │ │ │
        # * * * * *

        # All times are in UTC timezone

        # if self.routine == self.ROUTINE_OFTEN:  # for debug purposes
        #     return "* * * * *"

        if self.routine == self.ROUTINE_DAILY:
            return "0 5 * * *"

        if self.routine == self.ROUTINE_WEEKLY:
            return "0 5 * * 1"

        if self.routine == self.ROUTINE_MONTHLY:
            return "0 5 1 * *"

        if self.routine == self.ROUTINE_CUSTOM:
            return self.cron_string

        raise NotImplementedError()


auditlog.register(Export, exclude_fields=['modified', 'creator'])
auditlog.register(Scheduler, exclude_fields=['modified', 'creator', 'executions', 'job_id'])

from .signals import *



from django.contrib.auth import get_user_model
from django.core.mail import EmailMultiAlternatives
from django.db.models import Q
from django.utils import translation
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _
from django_rq import job
from whistle.helpers import notify


@job('exports')
def execute_export(exporter_class, exporter_params, language):
    # init exporter
    exporter = exporter_class(**exporter_params)

    # save export to DB
    export = exporter.save_export()
    
    # send mail with export to recipients
    export.send_mail(language, exporter_params.get('filename', None))


#  TODO: review
@job('exports')
def mail_export(export_id, export_class_name, language, filename=None):
    from outputs.models import Export
    export_class = import_string(export_class_name)
    export = export_class.objects.get(id=export_id)

    export.status = Export.STATUS_PROCESSING
    export.save(update_fields=['status'])

    # set language
    translation.activate(language)

    # model = self.queryset.model
    model = export.content_type.model_class()
    verbose_name = model._meta.verbose_name_plural

    # get exporter
    exporter = export.exporter

    # get total number of exported items
    num_items = export.total

    # export data to stream
    try:
        exporter.export()
    except Exception as e:
        # update status of export
        export.status = Export.STATUS_FAILED
        export.save(update_fields=['status'])

        # details
        details = '{}: {}\n'.format(_('Creator'), export.creator)
        details += '{}: {}\n\n'.format(_('Export ID'), str(export.id))
        details += '{}: {}\n'.format(_('Error'), str(e))

        notify_users = get_user_model().objects \
            .active()\
            .filter(
                Q(is_superuser=True) |
                Q(pk=export.creator.pk) |
                Q(pk__in=export.recipients.all())
            ).distinct()

        # notify creator, recipients and superusers about failed export
        for user in notify_users:
            notify(None, user, 'EXPORT_FAILED', object=export, target=export.content_type, details=details)

        raise

    if export.send_separately:
        for recipient in export.recipients_emails:
            message = get_message(
                exporter,
                count=num_items,
                recipient_list=[recipient],
                subject='{}: {}'.format(_('Export'), verbose_name),
                filename=filename
            )

            # send
            message.send(fail_silently=False)

    else:
        message = get_message(
            exporter,
            count=num_items,
            recipient_list=export.recipients_emails,
            subject='{}: {}'.format(_('Export'), verbose_name),
            filename=filename
        )
        # send
        message.send(fail_silently=False)

    # update status of export
    export.status = Export.STATUS_FINISHED
    export.save(update_fields=['status'])


def get_message(exporter, count, recipient_list, subject, filename=None):
    # message body
    body = exporter.get_message_body(count)

    # message subject
    subject = exporter.get_message_subject() or subject

    # prepare message
    message = EmailMultiAlternatives(subject=subject, to=recipient_list)
    message.attach_alternative(body, "text/html")

    if count > 0:
        # get the stream and set the correct mimetype
        message.attach(
            filename or exporter.get_filename(),
            exporter.get_output(),
            exporter.content_type
        )

    return message



from django.contrib import admin

from outputs.models import Export, Scheduler


@admin.register(Export)
class ExportAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    search_fields = ['creator__first_name', 'creator__last_name']
    list_select_related = ['creator', 'content_type']
    list_filter = ['format', 'context', 'content_type']
    list_display = ('id', 'content_type', 'format', 'creator', 'total', 'created')
    actions = ['send_mail']

    def send_mail(self, request, queryset):
        for obj in queryset.all():
            obj.send_mail(language=request.LANGUAGE_CODE)

    def view_on_site(self, obj):
        return obj.get_absolute_url()


@admin.register(Scheduler)
class SchedulerAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    search_fields = ['creator__first_name', 'creator__last_name']
    list_select_related = ['creator', 'content_type']
    list_filter = ['routine', 'is_active', 'format', 'context', 'content_type']
    list_display = ('id', 'is_active', 'routine', 'cron_string', 'cron_description', 'content_type', 'format', 'creator', 'created')




# Generated by Django 2.2.9 on 2020-08-21 11:21

from django.conf import settings
import django.contrib.postgres.fields
from django.db import migrations, models
import gm2m.fields


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='export',
            name='emails',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=33), blank=True, default=list, size=None),
        ),
        migrations.AddField(
            model_name='scheduler',
            name='emails',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=33), blank=True, default=list, size=None),
        ),
        migrations.AlterField(
            model_name='export',
            name='items',
            field=gm2m.fields.GM2MField(related_name='exports_where_item', through_fields=('gm2m_src', 'gm2m_tgt', 'gm2m_ct', 'gm2m_pk')),
        ),
        migrations.AlterField(
            model_name='export',
            name='recipients',
            field=models.ManyToManyField(blank=True, related_name='export_where_recipient', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='scheduler',
            name='recipients',
            field=models.ManyToManyField(blank=True, related_name='scheduler_where_recipient', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.2.9 on 2020-05-12 09:31

from django.conf import settings
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import gm2m.fields
from outputs import settings as outputs_settings


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    for dependency in outputs_settings.MIGRATION_DEPENDENCIES:
        dependencies.append(dependency)

    operations = [
        migrations.CreateModel(
            name='Scheduler',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('format', models.CharField(max_length=7)),
                ('context', models.CharField(max_length=10)),
                ('fields', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=40), blank=True, default=list, size=None)),
                ('query_string', models.TextField(blank=True, default='')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('routine', models.CharField(max_length=7)),
                ('is_active', models.BooleanField(default=True)),
                ('executions', django.contrib.postgres.fields.ArrayField(base_field=models.DateTimeField(), blank=True, default=list, size=None)),
                ('job_id', models.CharField(blank=True, max_length=36)),
                ('language', models.CharField(db_index=True, default='en', max_length=2)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('creator', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='scheduler_where_creator', to=settings.AUTH_USER_MODEL)),
                ('recipients', models.ManyToManyField(related_name='scheduler_where_recipient', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'scheduler',
                'verbose_name_plural': 'schedulers',
                'ordering': ('created',),
                'default_permissions': ('list', 'view', 'add', 'change', 'delete'),
            },
        ),
        migrations.CreateModel(
            name='Export',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('format', models.CharField(max_length=7)),
                ('context', models.CharField(max_length=10)),
                ('fields', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=40), blank=True, default=list, size=None)),
                ('query_string', models.TextField(blank=True, default='')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('status', models.CharField(default='PENDING', max_length=10)),
                ('total', models.PositiveIntegerField(default=0)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('creator', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='export_where_creator', to=settings.AUTH_USER_MODEL)),
                ('items', gm2m.fields.GM2MField(*outputs_settings.RELATED_MODELS, related_name='exports_where_item', through_fields=('gm2m_src', 'gm2m_tgt', 'gm2m_ct', 'gm2m_pk'))),
                ('recipients', models.ManyToManyField(related_name='export_where_recipient', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'export',
                'verbose_name_plural': 'exports',
                'ordering': ('created',),
                'default_permissions': ('list', 'view', 'add', 'change', 'delete'),
            },
        ),
    ]



import datetime
import io
import json
import operator

from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, QuerySet
from django.http import HttpResponse
from django.template import loader
from django.utils import translation
from django.utils.timezone import localtime
from django.utils.translation import ugettext_lazy as _

from pragmatic.templatetags.pragmatic_tags import filtered_values
from outputs import jobs
from outputs.forms import ChooseExportFieldsForm, ConfirmExportForm
from outputs.models import Export


class ExportFieldsPermissionsMixin(object):
    def load_export_fields_permissions(self, permissions):
        """
        Take single export_fields_permissions or their list or tuple and returns list of their json loaded versions, e.g. dictionaries
        """
        if isinstance(permissions, str):
            permissions = (permissions,)
        elif not isinstance(permissions, (list, tuple, QuerySet)):
            raise TypeError()

        loaded_permissions = []
        for permission in permissions:
            if not isinstance(permission, dict):
                permission = json.loads(permission)

            if isinstance(permission, str):
                permission = json.loads(permission)

            if not isinstance(permission, dict):
                raise TypeError()

            loaded_permissions.append(permission)

        return loaded_permissions

    def combine_export_fields_permissions(self, permissions):
        """
        Take list or tuple of json laoded export_fields_permissions and returns their union, e.g. combine them into a single dictionary
        """
        if not isinstance(permissions, (list, tuple, QuerySet)):
            raise TypeError()
        elif len(permissions) == 1:
            return permissions[0]

        result = {}
        for permission in permissions:
            for exporter, fields in permission.items():
                if exporter not in result:
                    result[exporter] = set()

                result[exporter].update(set(fields))

        for key, value in result.items():
            result[key] = list(value)

        return result

    def substract_export_fields_permissions(self, first, second):
        """
        Take two json laoded export_fields_permissions, first and second and returns first - second
        """
        if not isinstance(first, dict) or not isinstance(second, dict):
            raise TypeError()

        result = dict(first)
        for exporter, fields in first.items():
            if exporter in second:
                first_set = set(fields)
                second_set = set(second[exporter])
                if first_set <= second_set:
                    del result[exporter]
                else:
                    result[exporter] = list(set(fields) - set(second[exporter]))

        return result


class ConfirmExportMixin(object):
    template_name = 'outputs/export_confirmation.html'
    back_url = None
    form_class = ConfirmExportForm

    def get_success_url(self):
        return self.get_back_url()

    def get_initial(self):
        """
        Returns the initial data to use for forms on this view.
        """
        initial = super().get_initial()
        initial['recipients'] = self.request.user
        initial['filename'] = self.exporter_class.filename

        return initial

    @property
    def exporter_params(self):
        return {
            'user': self.request.user,
            'recipients': getattr(self, 'recipients', []),
            'params': self.get_params(),
            'filename': getattr(self, 'filename', None)
        }

    def get_params(self):
        params = self.request.GET.copy()

        try:
            params.pop('back_url')
        except KeyError:
            pass

        return params

    def get_back_url(self):
        url = self.request.GET.get('back_url', self.back_url)
        if not url:
            raise ValueError()
        return url

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data['back_url'] = self.get_back_url()
        context_data['objects_count'] = self.get_objects_count()
        return context_data

    def get_exporter(self):
        return self.exporter_class(**self.exporter_params)

    def export(self):
        jobs.execute_export.delay(self.exporter_class, self.exporter_params, language=translation.get_language())

    def get_objects_count(self):
        return self.get_exporter().get_queryset().count()

    def form_valid(self, form):
        self.recipients = form.cleaned_data.pop('recipients')
        self.filename = form.cleaned_data.pop('filename')
        self.export()
        messages.info(self.request, _('Your request is being processed. It can take few minutes. The result will be sent to your email.'))
        return super().form_valid(form)


class SelectExportMixin(ConfirmExportMixin, ExportFieldsPermissionsMixin):
    template_name = 'outputs/export_selection.html'
    form_class = ChooseExportFieldsForm
    limit = None
    selected_fields = []

    @property
    def exporter_params(self):
        return {
            'user': self.request.user,
            'recipients': getattr(self, 'recipients', []),
            'params': self.get_params(),
            'selected_fields': self.selected_fields,
            'filename': getattr(self, 'filename', None)
        }

    def form_valid(self, form):
        self.selected_fields = []
        for form_key, fields in form.cleaned_data.items():
            if form_key not in ['recipients', 'filename']:  # TODO: refactor using selectable_fields
                self.selected_fields += fields
        return super().form_valid(form)

    def get_permitted_fields(self):
        if not hasattr(self.exporter_class, 'selectable_fields'):
            return []

        permissions = []
        if self.request.user.export_fields_permissions:
            permissions.append(self.request.user.export_fields_permissions)

        permissions.extend(
            self.request.user.groups
                .exclude(metadata__export_fields_permissions__isnull=True)
                .values_list('metadata__export_fields_permissions', flat=True)
        )

        permissions = self.combine_export_fields_permissions(
            self.load_export_fields_permissions(permissions)
        )

        if not permissions:
            return []

        exporter_path = self.exporter_class.get_path()

        try:
            return permissions[exporter_path]
        except KeyError:
            return []

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()

        if hasattr(self.exporter_class, 'selectable_fields'):
            selectable_fields = self.exporter_class.selectable_fields()
        else:
            selectable_fields = None

        try:
            for set in self.exporter_class.selectable_iterative_sets().values():
                selectable_fields.update(set)
        except AttributeError:
            pass

        kwargs['selectable_fields'] = selectable_fields
        kwargs['permitted_fields'] = True if self.request.user.is_active and self.request.user.is_superuser else self.get_permitted_fields()
        return kwargs


class FilterExporterMixin(object):
    filter_class = None
    queryset = None
    model = None

    def __init__(self, params, **kwargs):
        self.params = params
        self.items = kwargs.pop('items', None)
        super().__init__(**kwargs)

        # create filter
        self.filter = self.get_filter()

    @classmethod
    def get_model(cls):
        return cls.queryset.model if cls.queryset is not None else cls.model

    @classmethod
    def get_app_and_model(cls):
        return cls.get_model()._meta.label.split('.')

    def get_filter(self):
        return self.filter_class(self.params, queryset=self.get_whole_queryset(self.params))

    def get_whole_queryset(self, params):
        proxy = params.get('proxy', None)

        if proxy:
            self.queryset = self.queryset.proxy(proxy)

        return self.queryset

    def get_queryset(self):
        items = getattr(self, 'items')

        if items:
            return self.filter.queryset.filter(pk__in=items)

        return self.filter.qs

    def get_message_body(self, count):
        template = loader.get_template('outputs/export_message_body.html')
        fv = filtered_values(self.filter, self.params)
        return template.render({'count': count, 'filtered_values': fv})


class ExporterMixin(object):
    content_type = 'application/force-download'
    filename = None
    export_format = None
    export_context = None
    send_separately = False
    description = ''

    def __init__(self, user, recipients, **kwargs):
        self.filename = kwargs.pop('filename', self.filename)
        self.send_separately = kwargs.pop('send_separately', self.send_separately)
        self.user = user
        self.recipients = recipients

        # initialize stream
        self.output = io.BytesIO()

    @classmethod
    def get_path(cls):
        return ".".join([cls.__module__, cls.__name__])

    @classmethod
    def get_description(cls):
        return cls.description

    def get_filename(self):
        if not self.filename:
            raise ValueError()

        # remove accents
        self.filename = self.filename.encode('ascii', 'ignore')
        self.filename = self.filename.decode('utf-8')

        return self.filename

    def export(self):
        raise NotImplementedError()

    def write_data(self, output):
        raise NotImplementedError()

    def get_output(self):
        self.output.seek(0)
        return self.output.read()

    def export_to_response(self):
        self.export()

        # construct response
        response = HttpResponse(
            self.get_output(),
            content_type=self.content_type,
        )
        response['Content-Disposition'] = "attachment; filename={}".format(
            self.get_filename()
        )

        return response

    def get_message_body(self, count):
        raise NotImplementedError()

    def get_message_subject(self):
        return None

    def save_export(self):
        items = self.get_queryset().all()
        model = self.queryset.model
        params = getattr(self, 'params', {})

        fields = getattr(self, 'selected_fields', None)

        if fields is None:  # exporting all fields by default (without fields selection)
            if hasattr(self, 'selectable_fields'):
                fields = []
                for label, fieldset in self.selectable_fields().items():
                    for field_definition in fieldset:
                        fields.append(field_definition[0])  # TODO: use map()
            else:
                # exporter doesn't have selectable fields (for instance export to PDF)
                # save None instead of empty list
                pass

        # track export
        export = Export.objects.create(
            content_type=ContentType.objects.get_for_model(model, for_concrete_model=False),
            format=self.export_format,
            context=self.export_context,
            exporter_path=".".join([self.__class__.__module__, self.__class__.__name__]),
            fields=fields,
            creator=self.user,
            query_string=params.urlencode(),
            total=items.count(),
            emails=[recipient.email for recipient in self.recipients]
        )
        export.recipients.add(*list(self.recipients))

        try:
            export.items.add(*list(items))
        except AttributeError:
            pass

        return export


class ExcelExporterMixin(ExporterMixin):
    content_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    export_format = Export.FORMAT_XLSX
    export_context = Export.CONTEXT_LIST
    FORMATS = {
        'bold': {'bold': True},
        'header': {'bold': True, 'font_color': '#ffffff', 'bg_color': '#E2105D'},
        'date': {'num_format': 'dd.mm.yyyy'},
        'datetime': {'num_format': 'dd.mm.yyyy hh:mm'},
        'time': {'num_format': 'hh:mm'},
        'integer': {'num_format': '#'},
        'money': {'num_format': '### ### ##0.00 €'},
        'bold_money': {'num_format': '### ### ##0.00 €', 'bold': True},
        'money_amount': {'num_format': '### ### ##0.00'},
        'bold_money_amount': {'num_format': '### ### ##0.00', 'bold': True},
    }
    proxy_class = None
    exclude_in_permission_widget = False

    @staticmethod
    def selectable_fields():
        raise NotImplementedError()

    def __init__(self, **kwargs):
        self.selected_fields = kwargs.get('selected_fields', None)
        super().__init__(**kwargs)

        # create a workbook in memory
        import xlsxwriter
        self.workbook = xlsxwriter.Workbook(self.output)
        self.workbook.remove_timezone = True

        # Add formats
        self.formats = {}
        for label, format in self.FORMATS.items():
            self.formats[label] = self.workbook.add_format(format)

    def get_worksheet_title(self, index=0):
        raise NotImplementedError()

    @staticmethod
    def to_excel_datetime(to_convert):
        from xlsxwriter.utility import datetime_to_excel_datetime
        return datetime_to_excel_datetime(to_convert, False, False)

    def export(self):
        self.write_data(self.workbook.add_worksheet(self.get_worksheet_title()))
        self.workbook.close()

    def get_attribute(self, field):
        return field[0]

    def get_label(self, field):
        return field[1]

    def get_column_width(self, field):
        return field[2]

    def get_cell_format(self, field):
        try:
            return field[3]
        except IndexError:
            return None

    def get_function(self, field):
        return field[4]

    def write_row(self, worksheet, row, col, obj, field):
        attr_index = None
        attr = self.get_attribute(field)

        if '[' in attr and ']' in attr:
            start = attr.rindex('[')
            end = attr.rindex(']')
            attr_index = attr[start+1:end]
            attr = attr[0:start]

        # get cell format
        cell_format_identifier = self.get_cell_format(field)
        cell_format = self.formats.get(cell_format_identifier, None)

        # get object attribute
        try:
            if isinstance(obj, dict):
                value = obj.get(attr)
            else:
                value = operator.attrgetter(attr)(obj)
        except AttributeError as e:
            if 'NoneType' in str(e):
                value = None
            else:
                raise e

        if attr_index and value is not None:
            value = value.get(attr_index, '')

        # try to use custom lambda handler
        try:
            func = self.get_function(field)
            value = func(value, obj)
        except TypeError:
            value = func(value)
        except IndexError:
            pass

        if isinstance(value, tuple) and isinstance(value[0], str):
            # formula
            formula, value = value
            worksheet.write_formula(row, col, formula, cell_format, value)
        else:
            # datetime
            if isinstance(value, datetime.datetime):
                value = localtime(value)

            if cell_format_identifier and 'date' in cell_format_identifier and value:
                # date or datetime format
                worksheet.write_datetime(row, col, value, cell_format)
            else:
                try:
                    # inherited string format
                    worksheet.write(row, col, value, cell_format)
                except TypeError:
                    # force string format
                    worksheet.write(row, col, str(value), cell_format)

    def get_queryset(self):
        raise NotImplementedError()

    def write_header(self, worksheet, fields, iterative_sets_fields):
        last_col = 0

        for field in fields:
            col_index = last_col
            attr = self.get_attribute(field)

            if hasattr(self, 'header_update') and attr in self.header_update.keys():
                label = self.header_update[attr]
            else:
                label = self.get_label(field)

            column_width = self.get_column_width(field)
            worksheet.write(0, col_index, label, self.formats['header'])
            worksheet.set_column(col_index, col_index, column_width)
            last_col += 1

        for iter_set in iterative_sets_fields:
            for iteration_index in range(1, iter_set['iteration_number'] + 1):
                for field in iter_set['fields']:
                    col_index = last_col
                    attr = self.get_attribute(field)
                    set_attr = iter_set['set_attr']

                    if hasattr(self, 'header_update') and set_attr in self.header_update.keys() and attr in self.header_update[set_attr]:
                        label = '{} #{}: {}'.format(iter_set['verbose_name'], iteration_index, self.header_update[set_attr][attr])
                    else:
                        label = '{} #{}: {}'.format(iter_set['verbose_name'], iteration_index, self.get_label(field))

                    worksheet.write(0, col_index, label, self.formats['header'])
                    worksheet.set_column(col_index, col_index, self.get_column_width(field))
                    last_col += 1

    def write_content(self, worksheet, fields, iterative_sets_fields, objects):
        # Write actual data. Start from the first cell. Rows and columns are zero indexed.
        row = 1
        max_col = 0

        for obj in objects:
            col = 0

            for field in fields:
                if self.proxy_class:
                    obj.__class__ = self.proxy_class
                self.write_row(worksheet, row, col, obj, field)
                col += 1

            for iter_set in iterative_sets_fields:
                for relative in getattr(obj, iter_set['set_attr']).all():
                    for field in iter_set['fields']:
                        self.write_row(worksheet, row, col, relative, field)
                        col += 1

            max_col = max(col, max_col)
            row += 1

        worksheet.autofilter(0, 0, row-1, max_col-1)

    def get_selected_fields(self, objects):
        fields = []

        for field_set in self.selectable_fields().values():
            for field in field_set:
                attr = self.get_attribute(field)
                if self.selected_fields is None or attr in self.selected_fields:
                    fields.append(field)

        iterative_sets_fields = []

        if hasattr(self, 'selectable_iterative_sets'):
            for set_attribute in self.selectable_iterative_sets().keys():
                # get number of iterations
                relatives_name = set_attribute[:-4]

                object_with_max_relatives = objects.annotate(
                    num_relatives=Count(relatives_name)
                ).order_by('-num_relatives')[0]
                iteration_number = object_with_max_relatives.num_relatives

                iterative_fields = []
                for field_set in self.selectable_iterative_sets()[set_attribute].values():
                    for field in field_set:
                        if self.get_attribute(field) in self.selected_fields:
                            iterative_fields.append(field)

                relatives_verbose_name = getattr(object_with_max_relatives, set_attribute).first()._meta.verbose_name

                iterative_sets_fields.append({
                    'set_attr': set_attribute,
                    'fields': iterative_fields,
                    'iteration_number': iteration_number,
                    'verbose_name': relatives_verbose_name,
                })

        return fields, iterative_sets_fields

    def write_data(self, worksheet):
        # get data
        objects = self.get_queryset()

        if not objects.exists():
            return

        # use only selected fields
        fields, iterative_sets_fields = self.get_selected_fields(objects)

        # write header and set columns width
        self.write_header(worksheet, fields, iterative_sets_fields)

        # write content
        self.write_content(worksheet, fields, iterative_sets_fields, objects)

        # write header and content for fields requiring iteration over multiple related objects if there are any
        # self.write_iterative_sets(worksheet, fields, objects)



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class OutputsConfig(AppConfig):
    name = 'outputs'
    verbose_name = _('Outputs')

    def schedule_jobs(self):
        from outputs.models import Scheduler

        for output_scheduler in Scheduler.objects.active():
            output_scheduler.schedule()



from django.db import migrations

from outputs.models import Export, Scheduler, AbstractExport


def set_exporter_path_export(*args, **kwargs):
    for export in Export.objects.all():
        export.exporter_path = AbstractExport.get_exporter_path(
            model_class=export.model_class,
            context=export.context,
            format=export.format)
        export.save(update_fields=['exporter_path'])


def set_exporter_path_scheduler(*args, **kwargs):
    for scheduler in Scheduler.objects.all():
        scheduler.exporter_path = AbstractExport.get_exporter_path(
            model_class=scheduler.model_class,
            context=scheduler.context,
            format=scheduler.format)
        scheduler.save(update_fields=['exporter_path'])


def unset_exporter_path_export(*args, **kwargs):
    Export.objects.update(exporter_path='')


def unset_exporter_path_scheduler(*args, **kwargs):
    Scheduler.objects.update(exporter_path='')


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0009_auto_20230111_1353'),
    ]

    operations = [
        migrations.RunPython(set_exporter_path_export, unset_exporter_path_export),
        migrations.RunPython(set_exporter_path_scheduler, unset_exporter_path_scheduler),
    ]



from django.conf import settings

EXCLUDE_EXPORTERS = getattr(settings, 'OUTPUTS_EXCLUDE_EXPORTERS', [])
EXPORTERS_MODULE_MAPPING = getattr(settings, 'OUTPUTS_EXPORTERS_MODULE_MAPPING', {})
MIGRATION_DEPENDENCIES = getattr(settings, 'OUTPUTS_MIGRATION_DEPENDENCIES', [])
RELATED_MODELS = getattr(settings, 'OUTPUTS_RELATED_MODELS', [])



# Generated by Django 3.1.14 on 2022-02-08 18:58

from django.conf import settings
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('outputs', '0006_auto_20200910_2149'),
    ]

    operations = [
        migrations.AlterField(
            model_name='export',
            name='fields',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=40), blank=True, default=None, null=True, size=None),
        )
    ]



# Generated by Django 3.1.14 on 2023-01-11 12:50

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0007_auto_20220208_1958'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scheduler',
            name='fields',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=40), blank=True, default=None, null=True, size=None),
        ),
    ]



# Generated by Django 2.2.9 on 2020-09-10 19:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0005_change_column_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='export',
            name='send_separately',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='scheduler',
            name='send_separately',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.1.10 on 2019-12-22 23:14
from django.db import migrations

from outputs.models import Export


def set_metadata(*args, **kwargs):
    print('Init emails in Exports ...')
    exports = Export.objects.all().only('emails')

    for export in exports:
        emails = list(export.recipients.values_list('email', flat=True))
        export.emails = emails

    print('>>> bulk update started ...')
    Export.objects.bulk_update(exports, fields=['emails'])
    print('>>> bulk update finished')


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0002_auto_20200821_1321'),
    ]

    operations = [
        migrations.RunPython(set_metadata, lambda *args, **kwargs: None)
    ]



from django.utils.module_loading import import_string
from django.utils.timezone import now

from outputs import jobs


# TODO: review
def schedule_export(scheduler_id, scheduler_class_name):
    # get scheduler by its identifier
    scheduler_class = import_string(scheduler_class_name)
    scheduler = scheduler_class.objects.get(pk=scheduler_id)

    # delay export job in background
    jobs.execute_export.delay(scheduler.exporter_class, scheduler.exporter_params, language=scheduler.language)

    # update list of execution datetimes
    scheduler.executions.append(now())
    scheduler.save(update_fields=['executions'])



# Generated by Django 3.1.14 on 2023-01-25 11:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('outputs', '0011_reformat_export_fields_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='scheduler',
            name='cron_string',
            field=models.CharField(blank=True, max_length=13),
        ),
        migrations.AddConstraint(
            model_name='scheduler',
            constraint=models.CheckConstraint(check=models.Q(('cron_string', ''), models.Q(_negated=True, routine='CUSTOM'), _connector='OR'), name='custom_cron_string'),
        )
    ]



from datetime import timedelta

import django_rq
from django.db.models.signals import post_save, pre_save, pre_delete
from django.dispatch import receiver
from whistle.helpers import notify

from django.contrib.auth import get_user_model
from outputs.models import Export, Scheduler
from outputs.signal_tasks import notify_about_executed_export, schedule_scheduler
from pragmatic.signals import SignalsHelper, apm_custom_context


@receiver(post_save, sender=Export)
@apm_custom_context('signals')
def export_executed_post_save(sender, instance, created, **kwargs):
    if created:
        scheduler = django_rq.get_scheduler('default')

        # schedule export notifications
        scheduler.enqueue_in(
            time_delta=timedelta(minutes=1),
            func=notify_about_executed_export,
            export=instance,
        )


@receiver(pre_save, sender=Scheduler)
@apm_custom_context('signals')
def reschedule_scheduler(sender, instance, **kwargs):
    """
    Signal to check if routine of the scheduler changed.
    """
    if SignalsHelper.attribute_changed(instance, ['is_active', 'routine']):
        SignalsHelper.add_task_and_connect(sender, instance, schedule_scheduler, [instance])


@receiver(pre_delete, sender=Scheduler)
@apm_custom_context('signals')
def cancel_scheduler(sender, instance, **kwargs):
    """
    Delete redis queue job when scheduler is deleted
    """
    instance.cancel_schedule()


@receiver(post_save, sender=Scheduler)
@apm_custom_context('signals')
def notify_about_scheduler(sender, instance, created, **kwargs):
    """
    Signal to notify when scheduler is created.
    """
    if created:
        recipients = get_user_model().objects.managers()

        if instance.creator:
            recipients = recipients.exclude(pk=instance.creator.pk)

        for recipient in recipients:
            notify(None, recipient, 'SCHEDULER_CREATED', actor=instance.creator, object=instance)



from django.contrib import messages
from django.shortcuts import get_object_or_404
from django.urls import reverse_lazy
from django.views.generic import ListView, CreateView, UpdateView, DetailView, DeleteView
from django.utils.translation import ugettext_lazy as _
from pragmatic.mixins import LoginPermissionRequiredMixin, DisplayListViewMixin, SortingListViewMixin, DeleteObjectMixin
from outputs.filters import ExportFilter, SchedulerFilter
from outputs.forms import SchedulerForm
from outputs.models import Export, Scheduler


class ExportListView(LoginPermissionRequiredMixin, DisplayListViewMixin, SortingListViewMixin, ListView):
    model = Export
    filter_class = ExportFilter
    permission_required = 'outputs.list_export'
    displays = ['table']
    paginate_values = [10, 50, 100]
    paginate_by_display = {'table': paginate_values}
    sorting_options = {'-created': _('Newest'), 'created': _('Oldest'), 'creator': _('Creator'), 'total': _('Total items')}

    def dispatch(self, request, *args, **kwargs):
        self.filter = self.filter_class(self.request.GET, queryset=self.get_whole_queryset())
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.sort_queryset(self.filter.qs)

    def get_whole_queryset(self):
        return self.model.objects.all()\
            .select_related('creator', 'content_type')\
            .prefetch_related('recipients')

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data.update({'filter': self.filter})
        return context_data


class SchedulerListView(LoginPermissionRequiredMixin, DisplayListViewMixin, SortingListViewMixin, ListView):
    model = Scheduler
    filter_class = SchedulerFilter
    permission_required = 'outputs.list_scheduler'
    displays = ['table']
    paginate_values = [10, 50, 100]
    paginate_by_display = {'table': paginate_values}
    sorting_options = {'-created': _('Newest'), 'created': _('Oldest'), 'creator': _('Creator')}

    def dispatch(self, request, *args, **kwargs):
        self.filter = self.filter_class(self.request.GET, queryset=self.get_whole_queryset())
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.sort_queryset(self.filter.qs)

    def get_whole_queryset(self):
        return self.model.objects.all() \
            .select_related('creator', 'content_type') \
            .prefetch_related('recipients')

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data.update({'filter': self.filter})
        return context_data


class SchedulerCreateView(LoginPermissionRequiredMixin, CreateView):
    model = Scheduler
    form_class = SchedulerForm
    permission_required = 'outputs.add_scheduler'

    def dispatch(self, request, *args, **kwargs):
        export_pk = kwargs.get('export_pk', None)
        self.export = get_object_or_404(Export, pk=export_pk) if export_pk else None
        return super().dispatch(request, *args, **kwargs)

    def get_initial(self):
        initial = super().get_initial()

        initial.update({
            'language': self.request.LANGUAGE_CODE
        })

        if self.export:
            initial.update({
                'content_type': self.export.content_type,
                'fields': self.export.fields,
                'format': self.export.format,
                'context': self.export.context,
                'exporter_path': self.export.exporter_path,
                'query_string': self.export.query_string,
                'recipients': self.export.recipients.all()
            })

        return initial

    def form_valid(self, form):
        messages.success(self.request, _('Scheduler successfully created'))
        scheduler = form.save(commit=False)
        scheduler.creator = self.request.user
        scheduler.save()
        return super().form_valid(form)


class SchedulerUpdateView(LoginPermissionRequiredMixin, UpdateView):
    model = Scheduler
    form_class = SchedulerForm
    permission_required = 'outputs.change_scheduler'

    def form_valid(self, form):
        messages.success(self.request, _('Scheduler successfully updated'))
        return super().form_valid(form)


class SchedulerDetailView(LoginPermissionRequiredMixin, DetailView):
    model = Scheduler
    permission_required = 'outputs.view_scheduler'


class SchedulerDeleteView(LoginPermissionRequiredMixin, DeleteObjectMixin, DeleteView):
    model = Scheduler
    success_url = reverse_lazy('outputs:scheduler_list')
    permission_required = 'outputs.delete_scheduler'



from crispy_forms.bootstrap import FormActions
from crispy_forms.layout import Layout, Row, Div, Fieldset, Submit
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.forms import BooleanField, CheckboxInput, TextInput, Textarea
from django.utils.translation import ugettext_lazy as _

from django.contrib.auth import get_user_model
from outputs.models import Scheduler, AbstractExport
from outputs.widgets import CheckboxSelectMultipleWithDisabled, Legend
from pragmatic.forms import SingleSubmitFormHelper


class ConfirmExportForm(forms.Form):
    recipients = forms.ModelMultipleChoiceField(
        label=_('Recipients'),
        queryset=get_user_model().objects.all(),
        # widget=UsersWidget(),
        required=True,
    )
    filename = forms.CharField(label=_('File name'))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = SingleSubmitFormHelper()
        self.helper.form_tag = False
        self.helper.layout = Layout(Row(
            Div('recipients', css_class='col-lg-3'),
            Div('filename', css_class='col-lg-3'),
        ))


class ChooseExportFieldsForm(ConfirmExportForm):
    select_all = BooleanField(label=_('Select all'), widget=CheckboxInput(attrs={'class': 'all'}), required=False)

    def __init__(self, *args, **kwargs):
        self.selectable_fields = kwargs.pop('selectable_fields')
        self.permitted_fields = kwargs.pop('permitted_fields', [])
        # necessary to reset dynamic fields, as they will be added again at page reload
        self.fields = {}

        super().__init__(*args, **kwargs)
        self.static_fields = list(self.fields.keys())

        self.prepare_fields()
        self.build_layouts()

        self.helper = SingleSubmitFormHelper()
        self.helper.form_tag = False
        self.helper.layout = Layout(*self.layouts)

    def prepare_fields(self):
        if self.selectable_fields:
            field_index = 0
            for label, field_group in self.selectable_fields.items():
                choices = []
                for field in field_group:
                    # disable choice if user doesn't have permission
                    if self.permitted_fields == True or field[0] in self.permitted_fields:
                        choices.append(field[:2])
                    else:
                        choices.append((field[0], {'label': field[1], 'disabled': True}))

                choices = tuple(choices)
                initial = tuple([field[0] for field in field_group if self.permitted_fields == True or field[0] in self.permitted_fields ])
                field_key = 'field_group_{}'.format(field_index)
                self.fields[field_key] = forms.MultipleChoiceField(label='', widget=CheckboxSelectMultipleWithDisabled(attrs={'class': 'option'}), choices=choices, initial=initial, required=False)

                # group checkbox
                group_key = 'group_{}'.format(field_index)
                self.fields[group_key] = BooleanField(label=label, widget=CheckboxInput(attrs={'class': 'group-option'}), required=False)

                field_index += 1

    def build_layouts(self):
        static_fields_layout = [
            Row(
                Div('recipients', css_class='col-lg-3'),
                Div('filename', css_class='col-lg-3'),
            ),
        ]

        # build dynamic fields layout
        dynamic_fields_layout = []
        for field_key, field in self.fields.items():
            # if not static field
            if field_key.startswith('field') and field_key not in self.static_fields:
                group_key = field_key.strip('field_')

                dynamic_fields_layout.append(
                    Div(
                        Legend(group_key),
                        Div(field_key),
                        css_class='col-md-2 group'
                    )
                )

        if dynamic_fields_layout:
            static_fields_layout.append(
                Row(
                    Div('select_all', css_class='col-md-3'),
                ),
            )

        self.layouts = [
            *static_fields_layout,
            Row(*dynamic_fields_layout),
        ]

    def clean(self):
        cleaned_data = super().clean()
        cleaned_data.pop('select_all')

        if self.selectable_fields:
            export_fields = []

            for key in list(cleaned_data.keys()):
                if key.startswith('group'):
                    cleaned_data.pop(key)
                elif key not in self.static_fields:
                    export_fields.append(cleaned_data[key])

            if not any(export_fields):
                raise forms.ValidationError(_('Select at least one option'))

        return cleaned_data


class SchedulerForm(forms.ModelForm):
    recipients = forms.ModelMultipleChoiceField(
        label=_('Recipients'),
        queryset=get_user_model().objects.all(),
        # widget=UsersWidget(),
        required=True,
    )

    content_type = forms.ModelChoiceField(label=_('content type'), queryset=ContentType.objects.order_by('model'))  # TODO: show only exportable content types

    class Meta:
        model = Scheduler
        fields = ('routine', 'cron_string', 'is_active', 'language',
                  'content_type', 'fields',
                  'format', 'context', 'exporter_path',
                  'query_string', 'recipients')
        widgets = {
            'query_string': TextInput,
            'fields': Textarea
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['query_string'].help_text = _('filter')

        self.helper = SingleSubmitFormHelper()
        self.helper.layout = Layout(
            Row(
                Fieldset(
                    _('Objects'),
                    Row(
                        Div('content_type', css_class='col-md-4'),
                        Div('context', css_class='col-md-4'),
                        Div('format', css_class='col-md-4'),
                        Div('exporter_path', css_class='col-md-12'),
                        Div('query_string', css_class='col-md-12'),
                        Div('fields', css_class='col-md-12'),
                    ),
                    css_class='col-md-4'
                ),
                Fieldset(
                    _('Management'),
                    'recipients',
                    'routine',
                    'cron_string',
                    'language',
                    'is_active',
                    css_class='col-md-2'
                )
            ),
            FormActions(
                Submit('submit', _('Submit'), css_class='btn-lg')
            )
        )

    def clean(self):
        context = self.cleaned_data.get('context', None)
        format = self.cleaned_data.get('format', None)
        content_type = self.cleaned_data.get('content_type', None)

        if context and format and content_type:
            model_class = content_type.model_class()

            try:
                # validation of content_type, context and format (check if exporter exists)
                AbstractExport.get_exporter_path(model_class, context, format)
            except ImportError:
                self.add_error(None, _('Exporter for content type {} with given context and format is not available').format(content_type))

        return self.cleaned_data
